"""Module for calculating wind speed drift."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from wind_up.plots.windspeed_drift_plots import plot_rolling_windspeed_diff_one_wtg
from wind_up.result_manager import result_manager

if TYPE_CHECKING:
    import pandas as pd

    from wind_up.models import PlotConfig, WindUpConfig


def _calculate_rolling_windspeed_diff(
    wtg_df: pd.DataFrame,
    *,
    ws_col: str,
    reanalysis_ws_col: str,
    timebase_s: int,
    ws_ll: float = 6,
    ws_ul: float = 15,
    rolling_period: float = 90,
    min_roll_days: float = 14,
    min_rolling_coverage: float = 1 / 3,
) -> pd.Series:
    ws_diff_to_renalysis = wtg_df[ws_col] - wtg_df[reanalysis_ws_col]
    ws_diff_to_renalysis.loc[(wtg_df[ws_col] < ws_ll) | (wtg_df[ws_col] > ws_ul)] = np.nan

    rows_per_day = 24 * 3600 / timebase_s

    def _rolling_specs(rolling_period: float) -> dict[str, int]:
        return {
            "window": round(rolling_period * rows_per_day),
            "min_periods": round(rolling_period * rows_per_day * min_rolling_coverage),
        }

    rolling_windspeed_diff = ws_diff_to_renalysis.rolling(**_rolling_specs(rolling_period)).median()

    while rolling_period >= (min_roll_days * 2) and len(rolling_windspeed_diff.dropna()) == 0:
        rolling_period = rolling_period // 2
        rolling_windspeed_diff = ws_diff_to_renalysis.rolling(**_rolling_specs(rolling_period)).median()

    if len(rolling_windspeed_diff.dropna()) == 0:
        result_manager.warning("could not calculate rolling windspeed diff")

    return rolling_windspeed_diff


def calc_max_abs_relative_rolling_windspeed_diff(ser: pd.Series) -> float:
    """Calculate the maximum absolute relative wind speed difference."""
    return (ser - ser.median()).abs().max()


def check_windspeed_drift(
    *,
    wtg_df: pd.DataFrame,
    wtg_name: str,
    ws_col: str,
    reanalysis_ws_col: str,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
    sub_dir: str | None = None,
) -> tuple[float, float]:
    """Calculate wind speed drift.

    :param wtg_df: time series of turbine measurement data
    :param wtg_name: name of the turbine
    :param ws_col: wind speed column name in `wtg_df`
    :param reanalysis_ws_col: reanalysis wind speed column name in `wtg_df`
    :param cfg: wind up configuration
    :param plot_cfg: plot configuration
    :param sub_dir: subdirectory in which to save plots
    :return:
        tuple containing

            - maximum absolute relative wind speed difference
            - maximum absolute relative wind speed drift during power performance period
    """
    rolling_windspeed_diff = _calculate_rolling_windspeed_diff(
        wtg_df, ws_col=ws_col, reanalysis_ws_col=reanalysis_ws_col, timebase_s=cfg.timebase_s
    )

    if plot_cfg is not None:
        plot_rolling_windspeed_diff_one_wtg(
            ser=rolling_windspeed_diff, wtg_name=wtg_name, ws_col=ws_col, plot_cfg=plot_cfg, sub_dir=sub_dir
        )

    max_abs_rel_diff = calc_max_abs_relative_rolling_windspeed_diff(rolling_windspeed_diff)
    max_abs_rel_diff_pp_period = calc_max_abs_relative_rolling_windspeed_diff(
        rolling_windspeed_diff.loc[cfg.analysis_first_dt_utc_start : cfg.analysis_last_dt_utc_start],  # type: ignore[misc]
    )

    ws_diff_ul = 1
    if max_abs_rel_diff > ws_diff_ul:
        result_manager.warning(f"possible wind speed drift of {max_abs_rel_diff:.1f} m/s for {wtg_name}")
    if max_abs_rel_diff_pp_period > ws_diff_ul:
        result_manager.warning(
            f"possible wind speed drift of {max_abs_rel_diff_pp_period:.1f} m/s for {wtg_name} "
            f"DURING POWER PERFORMANCE PERIOD",
        )

    return max_abs_rel_diff, max_abs_rel_diff_pp_period



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from wind_up.constants import REANALYSIS_WS_COL
from wind_up.models import WindUpConfig
from wind_up.windspeed_drift import _calculate_rolling_windspeed_diff, check_windspeed_drift


def test_check_windspeed_drift(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    test_name = "LSA_T13"
    test_ws_col = "test_ws_est_blend"
    test_max_ws_drift, test_max_ws_drift_pp_period = check_windspeed_drift(
        wtg_df=test_df,
        wtg_name=test_name,
        ws_col=test_ws_col,
        reanalysis_ws_col="test_" + REANALYSIS_WS_COL,
        cfg=cfg,
        plot_cfg=None,
    )
    assert test_max_ws_drift == pytest.approx(0.45289044075068974)
    assert test_max_ws_drift_pp_period == pytest.approx(0.42913942378401204)


def test_calc_rolling_windspeed_diff() -> None:
    n_values = 50
    timestep = pd.Timedelta("6h")
    ts_index = pd.date_range("2020-01-01", periods=n_values, freq=timestep)
    ws_col_vals = np.linspace(5, 15, n_values)
    test_df = pd.DataFrame({"ws_col": ws_col_vals, "reanalysis_ws_col": ws_col_vals[::-1]}, index=ts_index)

    original = test_df.copy()
    actual = _calculate_rolling_windspeed_diff(
        wtg_df=test_df,
        ws_col="ws_col",
        reanalysis_ws_col="reanalysis_ws_col",
        timebase_s=int(timestep / pd.Timedelta("1s")),
    )

    expected = pd.Series(np.nan, index=ts_index)
    expected[-17:] = np.linspace(-2.2448979591836746, 1.0204081632653068, 17)
    pd.testing.assert_series_equal(actual, expected)

    # checking original dataframe is not modified
    pd.testing.assert_frame_equal(test_df, original)



import timeit

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_series_equal

from tests.test_math_funcs import circ_median_exact
from wind_up.circular_math import circ_diff, rolling_circ_median_approx


@pytest.mark.parametrize("range_360", [True, False])
def test_rolling_circ_median(*, range_360: bool) -> None:
    timestamps = pd.date_range(start="2024-01-01", periods=8, freq="600s")
    input_df = pd.DataFrame(
        {
            "wind_direction_1": 2 * [359, 2.1, np.nan, 1],
            "wind_direction_2": 2 * [345, 30, np.nan, np.nan],
            "nacelle_direction": 2 * [359, 3, 358, 1],
        },
        index=timestamps,
    )
    window = 4
    min_periods = 3
    for col in input_df.columns:
        result = rolling_circ_median_approx(
            input_df[col], window=window, min_periods=min_periods, center=True, range_360=range_360
        )
        expected = input_df[col].rolling(window=window, min_periods=min_periods, center=True).apply(circ_median_exact)
        if not range_360:
            expected = ((expected + 180) % 360) - 180
        residuals = circ_diff(result, expected)
        if any(~residuals.isna()):
            assert residuals.abs().max() < 1e-1
        assert_series_equal(result, expected, atol=360)  # big atol to avoid wrap error


def test_rolling_circ_median_all_nans() -> None:
    timestamps = pd.date_range(start="2024-01-01", periods=8, freq="1s")
    input_df = pd.DataFrame(
        {"wind_direction": 8 * [np.nan], "nacelle_direction": 8 * [np.nan]},
        index=timestamps,
    )

    for col in input_df.columns:
        result = rolling_circ_median_approx(input_df[col], window=4, min_periods=3, center=True, range_360=True)
        expected = input_df[col].rolling(window=4, min_periods=3, center=True).apply(lambda x: circ_median_exact(x))
        assert_series_equal(result, expected)


@pytest.mark.slow
def test_rolling_circ_median_performance() -> None:
    # Generate a large dataset
    n_rows = 10_000
    n_cols = 1
    timestamps = pd.date_range(start="2024-01-01", periods=n_rows, freq="1s")

    rng = np.random.default_rng(0)
    data = np.concatenate([rng.normal(4, 20, n_rows // 2) % 360, rng.normal(354, 20, n_rows // 2) % 360])
    # Add some NaN values (5% of data)
    nan_rate = 0.05
    nan_mask = rng.random(size=data.shape) < nan_rate
    data[nan_mask] = np.nan

    input_df = pd.DataFrame(data, columns=[f"direction_{i}" for i in range(n_cols)], index=timestamps)

    window_size = 40
    min_periods = 30

    col = "direction_0"

    def new_method() -> float:
        return rolling_circ_median_approx(input_df[col], window=window_size, min_periods=min_periods, center=True)

    def exact_method() -> float:
        return (
            input_df[col]
            .rolling(window=window_size, min_periods=min_periods, center=True)
            .apply(lambda x: circ_median_exact(x))
        )

    new_method_results = new_method()
    exact_method_results = exact_method()
    residuals = circ_diff(new_method_results, exact_method_results)
    # results are not expected to exactly match
    assert abs(residuals.mean()) < 2e-2
    assert residuals.abs().mean() < 3e-1
    assert residuals.abs().max() < 6
    assert_series_equal(new_method_results, exact_method_results, atol=360)  # big atol to avoid wrap error

    # check speed
    number_of_runs = 10
    rolling_circ_median_time = timeit.timeit(new_method, number=number_of_runs)
    apply_exact_time = timeit.timeit(
        exact_method,
        number=number_of_runs,
    )
    minimum_speed_up = 100
    assert rolling_circ_median_time < apply_exact_time / minimum_speed_up, (
        f"New method ({rolling_circ_median_time:.2f}s) should be at least {minimum_speed_up}x faster than "
        f"old method ({apply_exact_time:.2f}s)"
    )



"""Backporting utilities for features in older Python versions."""

import sys
from collections.abc import Iterable, Iterator

if sys.version_info >= (3, 10):

    def strict_zip(*iterables: Iterable, strict: bool = False) -> Iterator:  # noqa: D103
        return zip(*iterables, strict=strict)
else:

    def strict_zip(*iterables: Iterable, strict: bool = False) -> Iterator:  # noqa: D103
        if strict:
            iterables = [list(it) for it in iterables]  # type: ignore[assignment]
            if not all(len(it) == len(iterables[0]) for it in iterables):  # type: ignore[arg-type]
                msg = "All iterables must have the same length"
                raise ValueError(msg)
        return zip(*iterables)  # type: ignore[call-overload]



"""Tests running the Smarteole dataset through wind-up analysis."""

from pathlib import Path

import pandas as pd
import pytest

from examples.smarteole_utils import SmartEoleExtractor
from tests.conftest import CACHE_DIR, TEST_DATA_FLD
from wind_up.interface import AssessmentInputs
from wind_up.main_analysis import run_wind_up_analysis
from wind_up.models import Asset, PlotConfig, Toggle, Turbine, WindUpConfig
from wind_up.reanalysis_data import ReanalysisDataset

SMARTEOLE_DATA_DIR = TEST_DATA_FLD / "smarteole"


def _define_wind_up_cfg(analysis_timebase_s: int, analysis_output_dir: Path) -> WindUpConfig:
    wtg_map = {
        f"SMV{i}": Turbine.model_validate(
            {
                "name": f"SMV{i}",
                "turbine_type": {
                    "turbine_type": "Senvion-MM82-2050",
                    "rotor_diameter_m": 82.0,
                    "rated_power_kw": 2050.0,
                    "cutout_ws_mps": 25,
                    "normal_operation_pitch_range": (-10.0, 35.0),
                    "normal_operation_genrpm_range": (250.0, 2000.0),
                    "rpm_v_pw_margin_factor": 0.05,
                    "pitch_to_stall": False,
                },
            }
        )
        for i in range(1, 7 + 1)
    }
    northing_corrections_utc = [
        ("SMV1", pd.Timestamp("2020-02-17 16:30:00+0000"), 5.750994540354649),
        ("SMV2", pd.Timestamp("2020-02-17 16:30:00+0000"), 5.690999999999994),
        ("SMV3", pd.Timestamp("2020-02-17 16:30:00+0000"), 5.558000000000042),
        ("SMV4", pd.Timestamp("2020-02-17 16:30:00+0000"), 5.936999999999996),
        ("SMV5", pd.Timestamp("2020-02-17 16:30:00+0000"), 6.797253350869262),
        ("SMV6", pd.Timestamp("2020-02-17 16:30:00+0000"), 5.030130916842758),
        ("SMV7", pd.Timestamp("2020-02-17 16:30:00+0000"), 4.605999999999972),
    ]

    wd_filter_margin = 3 + 7 * analysis_timebase_s / 600
    return WindUpConfig(
        assessment_name="smarteole_example",
        timebase_s=analysis_timebase_s,
        require_ref_wake_free=True,
        detrend_min_hours=12,
        ref_wd_filter=[207 - wd_filter_margin, 236 + wd_filter_margin],  # steer is from 207-236
        filter_all_test_wtgs_together=True,
        use_lt_distribution=False,
        out_dir=analysis_output_dir,
        test_wtgs=[wtg_map["SMV6"], wtg_map["SMV5"]],
        ref_wtgs=[wtg_map["SMV7"]],
        ref_super_wtgs=[],
        non_wtg_ref_names=[],
        analysis_first_dt_utc_start=pd.Timestamp("2020-02-17 16:30:00+0000"),
        upgrade_first_dt_utc_start=pd.Timestamp("2020-02-17 16:30:00+0000"),
        analysis_last_dt_utc_start=pd.Timestamp("2020-05-25 00:00:00+0000") - pd.Timedelta(seconds=analysis_timebase_s),
        lt_first_dt_utc_start=pd.Timestamp("2020-02-17 16:30:00+0000"),
        lt_last_dt_utc_start=pd.Timestamp("2020-05-25 00:00:00+0000") - pd.Timedelta(seconds=analysis_timebase_s),
        detrend_first_dt_utc_start=pd.Timestamp("2020-02-17 16:30:00+0000"),
        detrend_last_dt_utc_start=pd.Timestamp("2020-05-25 00:00:00+0000") - pd.Timedelta(seconds=analysis_timebase_s),
        years_for_lt_distribution=0,
        years_for_detrend=0,
        ws_bin_width=1.0,
        asset=Asset.model_validate(
            {
                "name": "Sole du Moulin Vieux",
                "wtgs": list(wtg_map.values()),
                "masts_and_lidars": [],
            }
        ),
        northing_corrections_utc=northing_corrections_utc,
        toggle=Toggle.model_validate(
            {
                "name": "wake steering",
                "toggle_file_per_turbine": False,
                "toggle_filename": "SMV_offset_active_toggle_df.parquet",
                "detrend_data_selection": "use_toggle_off_data",
                "pairing_filter_method": "any_within_timedelta",
                "pairing_filter_timedelta_seconds": 3600,
                "toggle_change_settling_filter_seconds": 120,
            }
        ),
    )


def _create_config(tmp_path: Path) -> AssessmentInputs:
    analysis_timebase_s = 600
    reanalysis_file_path = SMARTEOLE_DATA_DIR / "ERA5T_50.00N_2.75E_100m_1hr_20200201_20200531.parquet"

    input_data_cache_dir = CACHE_DIR / "test-smarteole"
    input_data_cache_dir.mkdir(exist_ok=True, parents=True)
    smarteole_extractor = SmartEoleExtractor(
        data_dir=SMARTEOLE_DATA_DIR,
        analysis_timebase_s=analysis_timebase_s,
    )
    scada_df = smarteole_extractor.unpack_smarteole_scada()
    metadata_df = smarteole_extractor.unpack_smarteole_metadata()
    toggle_df = smarteole_extractor.unpack_smarteole_toggle_data()

    cache_subdir = tmp_path / f"timebase_{analysis_timebase_s}"
    cache_subdir.mkdir(parents=True, exist_ok=True)

    # augment scada_df for plotting reasons
    toggle_df_no_tz = toggle_df.copy()
    toggle_df_no_tz.index = toggle_df_no_tz.index.tz_localize(None)
    scada_df = scada_df.merge(toggle_df_no_tz["yaw_offset_command"], left_index=True, right_index=True, how="left")
    scada_df["yaw_offset_command"] = scada_df["yaw_offset_command"].where(scada_df["TurbineName"] == "SMV6", 0)
    del toggle_df_no_tz

    cfg = _define_wind_up_cfg(analysis_timebase_s=analysis_timebase_s, analysis_output_dir=tmp_path)
    reanalysis_dataset = ReanalysisDataset(
        id="ERA5T_50.00N_2.75E_100m_1hr",
        data=pd.read_parquet(reanalysis_file_path),
    )

    plot_cfg = PlotConfig(show_plots=False, save_plots=True, plots_dir=cfg.out_dir / "plots")
    return AssessmentInputs.from_cfg(
        cfg=cfg,
        plot_cfg=plot_cfg,
        toggle_df=toggle_df,
        scada_df=scada_df,
        metadata_df=metadata_df,
        reanalysis_datasets=[reanalysis_dataset],
        cache_dir=cache_subdir,
    )


def _prep_results(input_df: pd.DataFrame, dp: int = 1) -> pd.DataFrame:
    input_df = input_df[
        [
            "test_wtg",
            "ref",
            "uplift_frc",
            "unc_one_sigma_frc",
            "uplift_p95_frc",
            "uplift_p5_frc",
            "pp_valid_hours_pre",
            "pp_valid_hours_post",
            "mean_power_post",
        ]
    ]
    for i, col in enumerate(x for x in input_df.columns if x.endswith("_frc")):
        if i == 0:
            output_df = input_df.assign(**{col: (input_df[col] * 100).round(dp).astype(str) + "%"})
        else:
            output_df = output_df.assign(**{col: (input_df[col] * 100).round(dp).astype(str) + "%"})
        output_df = output_df.rename(columns={col: col.replace("_frc", "_pct")})
    output_df["mean_power_post"] = output_df["mean_power_post"].round(0).astype("int64")
    return output_df.rename(
        columns={
            "test_wtg": "turbine",
            "ref": "reference",
            "uplift_pct": "energy uplift",
            "unc_one_sigma_pct": "uplift uncertainty",
            "uplift_p95_pct": "uplift P95",
            "uplift_p5_pct": "uplift P5",
            "pp_valid_hours_pre": "valid hours toggle off",
            "pp_valid_hours_post": "valid hours toggle on",
            "mean_power_post": "mean power toggle on",
        }
    )


@pytest.mark.slow
@pytest.mark.filterwarnings("ignore")
def test_smarteole_analysis(tmp_path: Path) -> None:
    """Test running the Smarteole analysis."""

    expected_print_df = pd.DataFrame(
        {
            "turbine": ["SMV6", "SMV5"],
            "reference": ["SMV7", "SMV7"],
            "energy uplift": ["-1.0%", "3.1%"],
            "uplift uncertainty": ["0.6%", "1.2%"],
            "uplift P95": ["-2.0%", "1.2%"],
            "uplift P5": ["-0.1%", "5.0%"],
            "valid hours toggle off": [132 + 3 / 6, 133 + 0 / 6],
            "valid hours toggle on": [136 + 0 / 6, 137 + 1 / 6],
            "mean power toggle on": [1148, 994],
        },
        index=[0, 1],
    )

    assessment_inputs = _create_config(tmp_path)
    results_per_test_ref_df = run_wind_up_analysis(assessment_inputs)
    actual_df = _prep_results(results_per_test_ref_df)

    pd.testing.assert_frame_equal(actual_df, expected_print_df)



from pathlib import Path

import pandas as pd
import pytest

from wind_up.circular_math import circ_median
from wind_up.constants import REANALYSIS_WD_COL
from wind_up.models import WindUpConfig
from wind_up.northing import _calc_max_abs_north_errs, apply_northing_corrections
from wind_up.scada_funcs import _scada_multi_index


def test_apply_northing_corrections(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    test_df.columns = test_df.columns.str.replace("test_", "")
    test_df["TurbineName"] = "LSA_T07"
    wf_df = test_df.copy()
    for wtg_name in ["LSA_T09", "LSA_T12", "LSA_T13", "LSA_T14"]:
        wtg_df = test_df.copy()
        wtg_df["TurbineName"] = wtg_name
        wf_df = pd.concat([wf_df, wtg_df])
    wf_df = _scada_multi_index(wf_df)
    wf_df_after_northing = apply_northing_corrections(wf_df, cfg=cfg, north_ref_wd_col=REANALYSIS_WD_COL, plot_cfg=None)

    median_yaw_before_northing = wf_df.groupby("TurbineName")["YawAngleMean"].apply(circ_median)
    median_yaw_after_northing = wf_df_after_northing.groupby("TurbineName")["YawAngleMean"].apply(circ_median)
    assert median_yaw_before_northing.min() == pytest.approx(237.45018005371094)
    assert median_yaw_before_northing.max() == pytest.approx(237.45018005371094)
    assert median_yaw_after_northing["LSA_T07"] == pytest.approx(222.450180)
    assert median_yaw_after_northing["LSA_T09"] == pytest.approx(253.450180)
    assert median_yaw_after_northing["LSA_T12"] == pytest.approx(77.712486)
    assert median_yaw_after_northing["LSA_T13"] == pytest.approx(240.450180)
    assert median_yaw_after_northing["LSA_T14"] == pytest.approx(228.450180)

    abs_north_errs_before_northing = _calc_max_abs_north_errs(
        wf_df, north_ref_wd_col=REANALYSIS_WD_COL, timebase_s=cfg.timebase_s
    )
    abs_north_errs_after_northing = _calc_max_abs_north_errs(
        wf_df_after_northing, north_ref_wd_col=REANALYSIS_WD_COL, timebase_s=cfg.timebase_s
    )
    assert abs_north_errs_before_northing.min() == pytest.approx(7.911393667045218)
    assert abs_north_errs_before_northing.max() == pytest.approx(7.911393667045218)
    assert abs_north_errs_after_northing["LSA_T07"] == pytest.approx(18.402401473162058)
    assert abs_north_errs_after_northing["LSA_T09"] == pytest.approx(174.10443861846687)
    assert abs_north_errs_after_northing["LSA_T12"] == pytest.approx(172.59341754958666)
    assert abs_north_errs_after_northing["LSA_T13"] == pytest.approx(10.894109266368332)
    assert abs_north_errs_after_northing["LSA_T14"] == pytest.approx(12.41110384502656)



"""WindUp configuration models."""

from __future__ import annotations

import datetime as dt
import json
import logging
from pathlib import Path
from typing import Annotated

import numpy as np
import pandas as pd
import yaml
from pydantic import BaseModel, BeforeValidator, Field, model_validator

from wind_up.backporting import strict_zip
from wind_up.constants import OUTPUT_DIR
from wind_up.yaml_loader import Loader, construct_include

logger = logging.getLogger(__name__)


class PlotConfig(BaseModel):
    """Plot configuration model."""

    show_plots: bool = Field(default=False, description="Show plots in interactive window")
    save_plots: bool = Field(default=True, description="Save plots to file")
    skip_per_turbine_plots: bool = Field(default=False, description="If True skip per turbine plots")
    plots_dir: Path = Field(description="Directory to save plots to")


class TurbineType(BaseModel):
    """Turbine type model."""

    turbine_type: str = Field(
        description="Turbine type description",
        min_length=2,
    )
    rotor_diameter_m: float = Field(description="Rotor diameter in meters", gt=0, examples=[100])
    rated_power_kw: float = Field(description="Rated power in kW", gt=0, examples=[3000])
    cutout_ws_mps: float = Field(default=25, description="Cutout wind speed in m/s", gt=0, examples=[25])
    normal_operation_pitch_range: tuple[float, float] = Field(
        description="Normal operation pitch range in degrees",
        examples=[[-10, 40]],
    )
    normal_operation_genrpm_range: tuple[float, float] = Field(
        description="Normal operation generator rpm range",
        examples=[[800, 1600]],
    )
    rpm_v_pw_margin_factor: float = Field(
        default=0.05,
        description="Factor to multiply max RPM by to get rpm margin for rpm vs power filtering",
        examples=[0.05, 0.1],
    )
    pitch_to_stall: bool = Field(
        default=False,
        description="If False then pitch filter will eliminate data below the threshold rather than above",
    )


def _validate_lat(value: float | None) -> float:
    v = np.nan if value is None else value
    min_value = -90
    max_value = 90
    if v < min_value or v > max_value:
        err_msg = f"Latitude {v} is out of range"
        raise ValueError(err_msg)
    return v


def _validate_lon(value: float | None) -> float:
    v = np.nan if value is None else value
    min_value = -180
    max_value = 180
    if v < min_value or v > max_value:
        err_msg = f"Latitude {v} is out of range"
        raise ValueError(err_msg)
    return v


class Turbine(BaseModel):
    """Turbine model."""

    name: str = Field(description="Turbine name", min_length=2)
    turbine_type: TurbineType
    latitude: Annotated[float, BeforeValidator(_validate_lat)] = np.nan
    longitude: Annotated[float, BeforeValidator(_validate_lon)] = np.nan

    def get_latlongs(self: Turbine) -> list[tuple[float, float]]:
        """Get the latitude and longitude of the turbine."""
        return [(self.latitude, self.longitude)]


class MastOrLidar(BaseModel):
    """Mast or LiDAR model."""

    name: str = Field(description="Object name", min_length=2)
    latitude: float = Field(default=np.nan, ge=-90, le=90)
    longitude: float = Field(default=np.nan, ge=-180, le=180)
    data_file_name: str = Field(
        min_length=2,
        description="Name of data timeseries file",
    )
    wind_speed_column: str = Field(min_length=2, description="Name of wind speed column in data timeseries file")
    wind_direction_column: str = Field(
        min_length=2,
        description="Name of wind direction column in data timeseries file",
    )


class Asset(BaseModel):
    """Asset model."""

    name: str = Field(description="Asset Name", min_length=2, examples=["Kelmarsh"])
    wtgs: list[Turbine]
    masts_and_lidars: list[MastOrLidar] = Field(description="list of mast and LiDAR objects, if any", default=[])


class Toggle(BaseModel):
    """Toggle model."""

    name: str = Field(description="Name of toggle signal to use in plots", min_length=2, default="toggle")
    toggle_file_per_turbine: bool = Field(
        description="Is there one toggle timeseries file per turbine, or one for the asset?",
    )
    toggle_filename: str = Field(
        min_length=2,
        description="Name of toggle timeseries file, can use {wtg} if toggle_file_per_turbine is True",
    )
    detrend_data_selection: str = Field(
        description="Method for selecting directional detrending data",
        default="do_not_use_toggle_test",
        examples=["do_not_use_toggle_test", "use_toggle_off_data"],
    )
    pairing_filter_method: str = Field(
        description="Method to use for pairing filter",
        examples=["none", "one_to_one", "any_within_timedelta"],
        default="none",
    )
    pairing_filter_timedelta_seconds: int = Field(
        description="Time delta in seconds to use for pairing filter",
        default=50 * 60,
    )
    toggle_change_settling_filter_seconds: int = Field(
        description="Time delta in seconds filter out after a toggle state change",
        default=10 * 60,
    )


class PrePost(BaseModel):
    """PrePost model."""

    pre_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in pre-upgrade analysis, UTC Start format",
    )
    pre_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in pre-upgrade analysis, UTC Start format",
    )
    post_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in post-upgrade analysis, UTC Start format",
    )
    post_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in post-upgrade analysis, UTC Start format",
    )


class WindUpConfig(BaseModel):
    """WindUpConfig model.

    This model is used to store the configuration for a WindUp analysis. The attributes form the basis for the analysis.
    """

    assessment_name: str = Field(
        min_length=2,
        description="Name used for assessment output folder",
    )
    timebase_s: int = Field(
        default=10 * 60,
        description="Timebase in seconds for SCADA data, other data is converted to this timebase",
    )
    ignore_turbine_anemometer_data: bool = Field(
        default=False,
        description="If true do not use turbine anemometer data for anything",
    )
    require_test_wake_free: bool = Field(
        default=False,
        description="Remove data where the test turbine has any upwind turbines",
    )
    require_ref_wake_free: bool = Field(
        default=False,
        description="Remove data where the reference has any upwind turbines",
    )
    detrend_min_hours: int = Field(
        default=24,
        description="Minimum number of hours to use in directional detrending",
    )
    ref_wd_filter: list[float] | None = Field(
        default=None,
        description="Wind direction filter for reference data; only data within this range is used",
        examples=[[195, 241]],
    )
    ref_hod_filter: list[float] | None = Field(
        default=None,
        description="Hour of day filter for reference data; only data within this range is used",
        examples=[[18, 6]],
    )
    filter_all_test_wtgs_together: bool = Field(
        description="If True any row filtered for one test wtg is filtered for all test wtgs. "
        "Recommended for wake steering analysis.",
        default=False,
    )
    use_lt_distribution: bool = Field(
        description="If True the long term distribution is calculated and used for uplift calculation", default=True
    )
    use_test_wtg_lt_distribution: bool = Field(
        description="If True the test turbine's data is used for the long term distribution, "
        "otherwise the whole wind farm is used",
        default=False,
    )
    out_dir: Path = Field(description="Directory to save assessment output to")
    test_wtgs: list[Turbine] = Field(description="List of test Turbine ids")
    ref_wtgs: list[Turbine] = Field(default=[], description="List of reference Turbine ids")
    ref_super_wtgs: list[str] = Field(
        default=[],
        description="List of reference super Turbine ids",
    )
    non_wtg_ref_names: list[str] = Field(
        default=[],
        description="List of non turbine references",
    )
    upgrade_first_dt_utc_start: dt.datetime = Field(
        description="First time when test turbine is upgraded, UTC Start format",
    )
    analysis_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in upgrade analysis, UTC Start format",
    )
    analysis_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in upgrade analysis, UTC Start format",
    )
    lt_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in long term analysis, UTC Start format",
    )
    lt_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in long term analysis, UTC Start format",
    )
    detrend_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in detrend analysis, UTC Start format",
    )
    detrend_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in detrend analysis, UTC Start format",
    )
    years_offset_for_pre_period: int | None = Field(
        description="How many years to go back in time to begin pre upgrade dataset",
        ge=0,
        examples=[None, 1],
        default=None,
    )
    years_for_lt_distribution: int = Field(
        description="How many years to use in long term analysis",
        ge=0,
        examples=[3, 5],
    )
    years_for_detrend: int = Field(
        description="How many years to use in directional detrending",
        ge=0,
        examples=[1, 2],
    )
    ws_bin_width: float = Field(description="Wind speed bin width in m/s", gt=0, examples=[0.5, 1])
    bootstrap_runs_override: int | None = Field(
        description="Number of bootstrap runs to use, if None then calculated",
        ge=0,
        examples=[None, 1000],
        default=None,
    )
    reanalysis_method: str = Field(
        default="node_with_best_ws_corr",
        description="Method to use for reanalysis (e.g. ERA5) selection",
        examples=["node_with_best_ws_corr"],
    )
    missing_scada_data_fields: list[str] = Field(
        default=[],
        description="List of SCADA fields not available for this asset",
        examples=[["YawAngleMin", "YawAngleMax"]],
    )
    asset: Asset
    exclusion_periods_utc: list[tuple[str, dt.datetime, dt.datetime]] = Field(
        default=[],
        description="Turbine id or 'ALL', first UTC timestamp to exclude, last UTC timestamp to exclude",
        examples=[
            "ALL",
            dt.datetime(2021, 9, 21, 12, 40, tzinfo=dt.timezone.utc),
            dt.datetime(2021, 9, 29, 18, 10, tzinfo=dt.timezone.utc),
        ],
    )
    yaw_data_exclusions_utc: list[tuple[str, dt.datetime, dt.datetime]] = Field(
        default=[],
        description="Turbine id or 'ALL', first UTC start format timestamp to exclude yaw data,"
        " last UTC timestamp to exclude yaw data",
        examples=[
            "ALL",
            dt.datetime(2021, 9, 21, 12, 40, tzinfo=dt.timezone.utc),
            dt.datetime(2021, 9, 29, 18, 10, tzinfo=dt.timezone.utc),
        ],
    )
    optimize_northing_corrections: bool = Field(
        default=False,
        description="run northing correction optimization",
    )
    northing_corrections_utc: list[tuple[str, dt.datetime, float]] = Field(
        default=[],
        description="Turbine id, first UTC timestamp to correct, correction to add to yaw data",
        examples=["SMV1", dt.datetime(2021, 9, 21, 12, 40, tzinfo=dt.timezone.utc), 15],
    )
    toggle: Toggle | None = None
    prepost: PrePost | None = None
    clip_rated_power_pp: bool = Field(default=True, description="Clip rated power in power performance analysis")
    use_rated_invalid_bins: bool = Field(default=False, description="Use rated power bins which have been filled in")

    @model_validator(mode="after")
    def _check_years_offset_for_pre_period(self: WindUpConfig) -> WindUpConfig:
        if self.toggle is None and self.years_offset_for_pre_period is None:
            msg = "toggle is None and years_offset_for_pre_period is None"
            raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def _check_first_datetime_before_last(self: WindUpConfig) -> WindUpConfig:
        if self.upgrade_first_dt_utc_start >= self.analysis_last_dt_utc_start:
            msg = "upgrade_first_datetime must be before last_useable_datetime"
            raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def _check_non_wtg_ref_names(self: WindUpConfig) -> WindUpConfig:
        for non_wtg_ref in self.non_wtg_ref_names:
            if non_wtg_ref == "reanalysis":
                if len(self.reanalysis_method) < 1:
                    msg = "reanalysis is in non_wtg_ref_names but reanalysis_method is not valid"
                    raise ValueError(msg)
            elif len([x for x in self.asset.masts_and_lidars if x.name == non_wtg_ref]) != 1:
                msg = f"there is not exactly 1 non_wtg_ref {non_wtg_ref} in cfg.asset.masts_and_lidars"
                raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def print_summary(self: WindUpConfig) -> WindUpConfig:
        """Print a summary of the WindUpConfig."""
        logger.info(f"loaded WindUpConfig assessment_name: {self.assessment_name}")
        dt_fmt = "%Y-%m-%d %H:%M"
        if self.toggle is not None:
            dt_rng_start = self.analysis_first_dt_utc_start.strftime(dt_fmt)
            dt_rng_end = (self.analysis_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
            logger.info(
                f"toggle analysis period (UTC): {dt_rng_start} to {dt_rng_end}",
            )
        elif self.prepost is not None:
            dt_rng_start = self.prepost.pre_first_dt_utc_start.strftime(dt_fmt)
            dt_rng_end = (self.prepost.pre_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
            logger.info(
                f"pre analysis period (UTC): {dt_rng_start} to {dt_rng_end}",
            )
            dt_rng_start = self.prepost.post_first_dt_utc_start.strftime(dt_fmt)
            dt_rng_end = (self.prepost.post_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
            logger.info(
                f"post analysis period (UTC): {dt_rng_start} to {dt_rng_end}",
            )
        else:
            msg = "toggle and prepost are both set to None"
            raise RuntimeError(msg)
        dt_rng_start = self.lt_first_dt_utc_start.strftime(dt_fmt)
        dt_rng_end = (self.lt_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
        logger.info(f"long term period (UTC): {dt_rng_start} to {dt_rng_end}")
        dt_rng_start = self.detrend_first_dt_utc_start.strftime(dt_fmt)
        dt_rng_end = (self.detrend_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
        logger.info(f"detrend period (UTC): {dt_rng_start} to {dt_rng_end}")
        return self

    @classmethod
    def from_yaml(cls, file_path: Path) -> "WindUpConfig":  # noqa ANN102
        """Load WindUpConfig from a yaml file.

        :param file_path: Path to yaml file
        :return: WindUpConfig
        """
        yaml.add_constructor("!include", construct_include, Loader)
        with Path.open(file_path) as f:
            cfg_dct = yaml.load(f, Loader)  # noqa S506

        cfg_dct["out_dir"] = OUTPUT_DIR / cfg_dct["assessment_name"]

        # ensure datetimes are UTC aware
        cfg_dct["upgrade_first_dt_utc_start"] = pd.Timestamp(cfg_dct["upgrade_first_dt_utc_start"], tz="UTC")
        cfg_dct["analysis_last_dt_utc_start"] = pd.Timestamp(cfg_dct["analysis_last_dt_utc_start"], tz="UTC")
        if cfg_dct.get("exclusion_periods_utc", None) is not None:
            for x in cfg_dct["exclusion_periods_utc"]:
                x[1] = pd.Timestamp(x[1], tz="UTC")
                x[2] = pd.Timestamp(x[2], tz="UTC")
        if cfg_dct.get("yaw_data_exclusions_utc", None) is not None:
            for x in cfg_dct["yaw_data_exclusions_utc"]:
                x[1] = pd.Timestamp(x[1], tz="UTC")
                x[2] = pd.Timestamp(x[2], tz="UTC")
        if cfg_dct.get("northing_corrections_utc", None) is not None:
            for x in cfg_dct["northing_corrections_utc"]:
                x[1] = pd.Timestamp(x[1], tz="UTC")

        # calculate analysis and long term datetimes
        test_is_toggle = cfg_dct.get("toggle", None) is not None
        if test_is_toggle:
            cfg_dct["analysis_first_dt_utc_start"] = cfg_dct["upgrade_first_dt_utc_start"]
        else:
            pre_post_dict = {
                "post_first_dt_utc_start": cfg_dct["upgrade_first_dt_utc_start"],
                "post_last_dt_utc_start": cfg_dct["analysis_last_dt_utc_start"],
                "pre_first_dt_utc_start": cfg_dct["upgrade_first_dt_utc_start"]
                - pd.DateOffset(years=cfg_dct["years_offset_for_pre_period"]),
                "pre_last_dt_utc_start": cfg_dct["analysis_last_dt_utc_start"]
                - pd.DateOffset(years=cfg_dct["years_offset_for_pre_period"]),
            }
            cfg_dct["prepost"] = PrePost.model_validate(pre_post_dict)
            cfg_dct["analysis_first_dt_utc_start"] = cfg_dct["prepost"].pre_first_dt_utc_start
        cfg_dct["lt_last_dt_utc_start"] = (
            cfg_dct["upgrade_first_dt_utc_start"] - dt.timedelta(days=7)  # go back 1 week for buffer before toggling
            if test_is_toggle
            else cfg_dct["prepost"].pre_last_dt_utc_start
        )
        cfg_dct["lt_first_dt_utc_start"] = cfg_dct["lt_last_dt_utc_start"] - pd.DateOffset(
            years=cfg_dct["years_for_lt_distribution"]
        )
        cfg_dct["detrend_last_dt_utc_start"] = cfg_dct["lt_last_dt_utc_start"]
        if test_is_toggle and cfg_dct["toggle"].get("detrend_data_selection", None) == "use_toggle_off_data":
            cfg_dct["detrend_last_dt_utc_start"] = cfg_dct["analysis_last_dt_utc_start"]
        cfg_dct["detrend_first_dt_utc_start"] = cfg_dct["detrend_last_dt_utc_start"] - pd.DateOffset(
            years=cfg_dct["years_for_detrend"]
        )

        # check each test_wtg exists in asset.wtgs
        for x in cfg_dct["test_wtgs"]:
            if x not in cfg_dct["asset"]["wtgs"]:
                msg = f"test_wtg {x} does not exist in asset.wtgs"
                raise ValueError(msg)

        # if ref_wtgs exist, check each ref_wtg exists in asset.wtgs
        if cfg_dct.get("ref_wtgs", None) is not None:
            for x in cfg_dct["ref_wtgs"]:
                if x not in cfg_dct["asset"]["wtgs"]:
                    msg = f"ref_wtg {x} does not exist in asset.wtgs"
                    raise ValueError(msg)

        # change lists of strings into list of turbines
        if len(cfg_dct["asset"]["turbine_types"]) == 1:
            cfg_dct["asset"]["turbine_types"] = cfg_dct["asset"]["turbine_types"] * len(
                cfg_dct["asset"]["wtgs"],
            )
        tt_list = [TurbineType.model_validate(tt) for tt in cfg_dct["asset"]["turbine_types"]]
        cfg_dct["asset"]["wtgs"] = [
            Turbine(name=x, turbine_type=tt) for x, tt in strict_zip(cfg_dct["asset"]["wtgs"], tt_list)
        ]
        test_wtg_list = []
        for x in cfg_dct["test_wtgs"]:
            for wtg in cfg_dct["asset"]["wtgs"]:
                if x == wtg.name:
                    test_wtg_list.append(wtg)
                    continue
        cfg_dct["test_wtgs"] = test_wtg_list
        if cfg_dct.get("ref_wtgs", None) is not None:
            ref_wtg_list = []
            for x in cfg_dct["ref_wtgs"]:
                for wtg in cfg_dct["asset"]["wtgs"]:
                    if x == wtg.name:
                        ref_wtg_list.append(wtg)
                        continue
            cfg_dct["ref_wtgs"] = ref_wtg_list
        elif cfg_dct.get("non_wtg_ref_names", None) is not None:
            if len(cfg_dct["non_wtg_ref_names"]) < 1:
                msg = "ref_wtgs or non_wtg_ref_names must have more than one element"
                raise ValueError(msg)
        else:
            msg = "ref_wtgs or non_wtg_ref_names must be specified"
            raise ValueError(msg)
        return WindUpConfig.model_validate(cfg_dct)

    def save_json(self, file_path: Path | None = None) -> None:
        """Save WindUpConfig to a json file.

        :param file_path: Path to json file
        :return: None
        """
        file_path = file_path or (self.out_dir / f"config_{self.assessment_name}.json")
        if not file_path.name.endswith(".json"):
            err_msg = "file_path must end with .json"
            raise ValueError(err_msg)
        if not file_path.parent.exists():
            file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open(mode="w") as f:
            json_data = json.loads(self.model_dump_json().replace("NaN", "null"))  # required to serialize NaN
            json.dump(json_data, f, indent=4)
        logger.info(f"Saved WindUpConfig to {file_path}")

    def get_max_rated_power(self: WindUpConfig) -> float:
        """Get the maximum rated power out of all turbine types on the wind farm asset."""
        return max(x.turbine_type.rated_power_kw for x in self.asset.wtgs)

    def list_unique_turbine_types(self: WindUpConfig) -> list[TurbineType]:
        """List unique turbine types of the wind farm asset."""
        unique_names = sorted({x.turbine_type.turbine_type for x in self.asset.wtgs})
        unique_turbine_types = []
        for name in unique_names:
            unique_turbine_types.append(
                next(x.turbine_type for x in self.asset.wtgs if x.turbine_type.turbine_type == name),
            )
        return unique_turbine_types

    def list_turbine_ids_of_type(self: WindUpConfig, ttype: TurbineType) -> list[str]:
        """List turbine ids of a specific turbine type.

        :param ttype: TurbineType
        :return: list of turbine ids
        """
        return [x.name for x in self.asset.wtgs if x.turbine_type == ttype]

    def get_normal_operation_genrpm_range(self: WindUpConfig, ttype: TurbineType) -> tuple[float, float]:
        """Get the normal operation generator rpm range for turbine type.

        :param ttype: TurbineType
        :return: tuple of generator rpm range
        """
        return next(x.turbine_type.normal_operation_genrpm_range for x in self.asset.wtgs if x.turbine_type == ttype)

    def get_normal_operation_pitch_range(self: WindUpConfig, ttype: TurbineType) -> tuple[float, float]:
        """Get the normal operation pitch range for turbine type.

        :param ttype: TurbineType
        :return: tuple of pitch range
        """
        return next(x.turbine_type.normal_operation_pitch_range for x in self.asset.wtgs if x.turbine_type == ttype)



"""Tests for the `WeDoWind` example wind farm."""

import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import wind_up
from examples.helpers import setup_logger
from examples.wedowind_example import main_wedowind_analysis
from tests.conftest import TEST_DATA_FLD


@pytest.fixture
def wedowind_expected_pitch_angle_results() -> pd.DataFrame:
    shutil.rmtree(TEST_DATA_FLD / "wedowind/Pitch Angle", ignore_errors=True)
    return pd.DataFrame(
        [
            {
                "preprocess_warning_counts": 0,
                "test_warning_counts": 7,
                "test_ref_warning_counts": 4,
                "test_wtg": "Test",
                "test_pw_col": "test_ActivePowerMean",
                "ref": "Ref",
                "ref_ws_col": "ref_ws_est_blend",
                "uplift_frc": 0.030473380515171976,
                "unc_one_sigma_frc": 0.006068651042272032,
                "uplift_p95_frc": 0.02049133783758799,
                "uplift_p5_frc": 0.04045542319275596,
                "pp_data_coverage": 0.07030065689742294,
                "distance_m": 1727.532358659697,
                "bearing_deg": 80.01879303945559,
                "unc_one_sigma_noadj_frc": 0.004194965381090924,
                "unc_one_sigma_lowerbound_frc": 0.0043787000647438,
                "unc_one_sigma_bootstrap_frc": 0.006068651042272032,
                "ref_max_northing_error_v_reanalysis": np.nan,
                "ref_max_northing_error_v_wf": np.nan,
                "ref_max_ws_drift": np.nan,
                "ref_max_ws_drift_pp_period": np.nan,
                "ref_powercurve_shift": -0.03498701513899527,
                "ref_rpm_shift": 0.0,
                "ref_pitch_shift": 0.0,
                "detrend_pre_r2_improvement": 6.304604371898392e-05,
                "detrend_post_r2_improvement": 5.719700032302821e-05,
                "mean_power_pre": 827.0693190786338,
                "mean_power_post": 884.7969322535646,
                "mean_test_yaw_offset_pre": 0.0,
                "mean_test_yaw_offset_post": 0.0,
                "t_value_one_sigma": 1.000403388430281,
                "missing_bins_unc_scale_factor": 1.0001853241828869,
                "pp_valid_hours_pre": 206.83333333333334,
                "pp_valid_hours_post": 349.66666666666663,
                "pp_valid_hours": 556.5,
                "pp_invalid_bin_count": 8,
                "uplift_noadj_frc": 0.034852080579915776,
                "poweronly_uplift_frc": 0.03616453743718513,
                "reversed_uplift_frc": 0.027407137307697532,
                "reversal_error": -0.0087574001294876,
                "lt_wtg_hours_raw": 3536.3333333333335,
                "lt_wtg_hours_filt": 3536.333333333333,
                "test_max_ws_drift": np.nan,
                "test_max_ws_drift_pp_period": np.nan,
                "test_powercurve_shift": 0.0861442820830347,
                "test_rpm_shift": 0.0,
                "test_pitch_shift": 0.0,
            }
        ],
        index=[0],
    )


@pytest.fixture
def wedowind_expected_vg_results() -> pd.DataFrame:
    shutil.rmtree(TEST_DATA_FLD / "wedowind/Vortex Generator", ignore_errors=True)
    return pd.DataFrame(
        [
            {
                "preprocess_warning_counts": 0,
                "test_warning_counts": 3,
                "test_ref_warning_counts": 3,
                "test_wtg": "Test",
                "test_pw_col": "test_pw_clipped",
                "ref": "Ref",
                "ref_ws_col": "ref_ws_est_blend",
                "uplift_frc": 0.023046823561716365,
                "unc_one_sigma_frc": 0.005119990647449296,
                "uplift_p95_frc": 0.014625188375301775,
                "uplift_p5_frc": 0.03146845874813096,
                "pp_data_coverage": 0.0736418776371308,
                "distance_m": 1727.532358659697,
                "bearing_deg": 80.01879303945559,
                "unc_one_sigma_noadj_frc": 0.004255707445544146,
                "unc_one_sigma_lowerbound_frc": 0.0020446262880886657,
                "unc_one_sigma_bootstrap_frc": 0.005119990647449296,
                "ref_max_northing_error_v_reanalysis": np.nan,
                "ref_max_northing_error_v_wf": np.nan,
                "ref_max_ws_drift": np.nan,
                "ref_max_ws_drift_pp_period": np.nan,
                "ref_powercurve_shift": 0.005226755665862948,
                "ref_rpm_shift": 0.0,
                "ref_pitch_shift": 0.0,
                "detrend_pre_r2_improvement": 9.392745002734237e-05,
                "detrend_post_r2_improvement": 0.00013574805792260225,
                "mean_power_pre": 456.00674788268043,
                "mean_power_post": 492.0387530327367,
                "mean_test_yaw_offset_pre": 0.0,
                "mean_test_yaw_offset_post": 0.0,
                "t_value_one_sigma": 1.0002396357467755,
                "missing_bins_unc_scale_factor": 1.030719576586684,
                "pp_valid_hours_pre": 396.66666666666663,
                "pp_valid_hours_post": 348.0,
                "pp_valid_hours": 744.6666666666666,
                "pp_invalid_bin_count": 12,
                "uplift_noadj_frc": 0.0210021972736277,
                "poweronly_uplift_frc": 0.01271095866103132,
                "reversed_uplift_frc": 0.01680021123720865,
                "reversal_error": 0.004089252576177331,
                "lt_wtg_hours_raw": 3498.6666666666665,
                "lt_wtg_hours_filt": 3498.6666666666665,
                "test_max_ws_drift": np.nan,
                "test_max_ws_drift_pp_period": np.nan,
                "test_powercurve_shift": 0.017597853713183786,
                "test_rpm_shift": 0.0,
                "test_pitch_shift": 0.0,
            }
        ],
        index=[0],
    )


@pytest.mark.slow
@pytest.mark.filterwarnings("ignore")
def test_wedowind_pitch_angle(wedowind_expected_pitch_angle_results: pd.DataFrame, tmp_path: Path) -> None:
    """Test the WeDoWind Pitch Angle example."""
    cache_dir = TEST_DATA_FLD / "wedowind"
    setup_logger(tmp_path / "analysis.log")
    actual = main_wedowind_analysis(
        analysis_name="Pitch Angle",
        generate_custom_plots=False,
        save_plots=False,
        cache_dir=cache_dir,
        analysis_output_dir=tmp_path,
        bootstrap_runs_override=40,  # to speed up test
    )
    assert actual["wind_up_version"].squeeze() == wind_up.__version__
    pd.testing.assert_frame_equal(
        actual.drop(columns=["time_calculated", "wind_up_version"]), wedowind_expected_pitch_angle_results
    )

    shutil.rmtree(cache_dir / "Pitch Angle", ignore_errors=True)  # clean up cache


@pytest.mark.slow
@pytest.mark.filterwarnings("ignore")
def test_wedowind_vortex_generators(wedowind_expected_vg_results: pd.DataFrame, tmp_path: Path) -> None:
    """Test the WeDoWind Vortex Generators example."""
    cache_dir = TEST_DATA_FLD / "wedowind"
    setup_logger(tmp_path / "analysis.log")
    actual = main_wedowind_analysis(
        analysis_name="Vortex Generator",
        generate_custom_plots=False,
        save_plots=False,
        cache_dir=cache_dir,
        analysis_output_dir=tmp_path,
        bootstrap_runs_override=40,  # to speed up test
    )
    assert actual["wind_up_version"].squeeze() == wind_up.__version__
    pd.testing.assert_frame_equal(
        actual.drop(columns=["time_calculated", "wind_up_version"]), wedowind_expected_vg_results
    )

    shutil.rmtree(cache_dir / "Vortex Generator", ignore_errors=True)  # clean up
